module Cli where

import RIO
import Data.UUID

import Options.Applicative

import ExitCodes


data Command = Initialize String
             | SyncOne String
             | ApplyOne UUID
             | ApplyMany String
             | SendSince UUID Integer
             | Purge
             | PrintUUID
             | ShowGlobalState
             | EncodeGlobalState
             | RequestFrom String
             | InitialSyncForPeer UUID String
             | InitialSyncFrom
             | SyncAllMailBoxes
             | SyncAllMailBoxesInit
             | Update String
             | UpdateAll
             | Uux UUID String [String]
             | DropPeer UUID
  deriving (Show, Read, Eq)

data Options = Options
    { cfgFile :: String
    , optInitialLog :: Bool
    , optCommand :: Command
    } deriving (Show, Read, Eq)

opts :: Parser Options
opts = Options <$> strOption ( long "config" <> short 'c' <> metavar "CONFIG" <> help "path to config file"
                              <> showDefault <> value "~/.mail-syncrc")
               <*> switch (long "initial-log" <> help "Also log events before the location of the log file is determined")
  <*> subparser
  (command "setup" (info ((subparser
    (   command "init" (info initOptions (progDesc "Initialize this peer"))
     <> command "printUUID" (info printUUIDOptions (progDesc "Show own UUID"))
     <> command "syncAllMailBoxesInit" (info syncAllMailBoxesInitOptions
                                             (progDesc "Sync all mailboxes initially (only first node of the network)"))
     <> command "initialSyncFor" (info initialSyncForOptions (progDesc "Request initial sync for the given peer"))
     <> command "initialSyncFrom" (info initialSyncFromOptions (progDesc "Accepts initial sync"))
     <> command "dropPeer" (info dropPeerOptions (progDesc "Drop a peer from your network"))
                          ) <**> helper)) (fullDesc <> progDesc "Commands for the set up"))
  <> command "peerToPeerProtocol" (info ((subparser
     (    command "requestFrom" (info requestOptions (progDesc "Request work from the given peer"))
       <> command "sendSince" (info sendSinceOptions (progDesc "Send all changes since"))
       <> command "applyMany" (info applyManyOptions (progDesc "Apply many changes (binary)"))
                          ) <**> helper)) (fullDesc <> progDesc "Commands for generating or processing messages of the peer-to-peer protocol"))
  <> command "debugState" (info ((subparser
    (   command "showGlobalState" (info showGlobalStateOptions (progDesc "Show own the global State (non-binary)"))
     <> command "encodeGlobalState" (info encodeGlobalStateOptions (progDesc "Encode a non-binary global state to binary"))
                          ) <**> helper)) (fullDesc <> progDesc "Commands for debugging the internal state"))
  <> command "syncOne" (info syncOneOptions (progDesc "Sync one mailbox locally"))
  <> command "syncAllMailBoxes" (info syncAllMailBoxesOptions (progDesc "Sync all mailboxes"))
  <> command "uux" (info uuxOptions (progDesc "Schedule the execution of a command on a peer"))
  <> command "update" (info updateOptions (progDesc "Updates with a peer"))
  <> command "updateAll" (info updateAllOptions (progDesc "Updates with all peers"))
  <> command "purge" (info purgeOptions (progDesc "Purge unneeded entries in the global state"))
  )

optsWithInfo :: ParserInfo Options
optsWithInfo = info (opts <**> helper)
    ( fullDesc
   <> progDesc "Syncronize Mails"
   <> failureCode exitFailedToParseCmdLine)

initOptions :: Parser Command
initOptions = Initialize <$> argument str (metavar "NAME" <> help "Symbolic name of this node")

syncOneOptions :: Parser Command
syncOneOptions = SyncOne <$> argument str (metavar "MAILBOX" <> help "Mailbox to sync")

applyOneOptions :: Parser Command
applyOneOptions = ApplyOne <$> argument auto (metavar "PEER-ID" <> help "Peer who sends the patch")

applyManyOptions :: Parser Command
applyManyOptions = ApplyMany <$> argument str (metavar "PEER-ID" <> help "Peer who sends the patch")

sendSinceOptions :: Parser Command
sendSinceOptions = liftA2 SendSince (argument auto (metavar "PEER-ID" <> help "Peer who requests the updates")) 
  $ argument auto (metavar "SINCE" <> help "First change to send, peer knows all previous changes")

purgeOptions :: Parser Command
purgeOptions = pure Purge

printUUIDOptions :: Parser Command
printUUIDOptions = pure PrintUUID

showGlobalStateOptions :: Parser Command
showGlobalStateOptions = pure ShowGlobalState

encodeGlobalStateOptions :: Parser Command
encodeGlobalStateOptions = pure EncodeGlobalState

requestOptions :: Parser Command
requestOptions = RequestFrom <$> argument str (metavar "PEER-ID" <> help "Peer to request updates from")

initialSyncForOptions :: Parser Command
initialSyncForOptions = liftA2 InitialSyncForPeer (argument auto (metavar "PEER-ID" <> help "Peer to request initial sync for")) $ argument str (metavar "PEER-NAME" <> help "Name of peer")

initialSyncFromOptions :: Parser Command
initialSyncFromOptions = pure InitialSyncFrom

syncAllMailBoxesOptions :: Parser Command
syncAllMailBoxesOptions = pure SyncAllMailBoxes

syncAllMailBoxesInitOptions :: Parser Command
syncAllMailBoxesInitOptions = pure SyncAllMailBoxesInit

updateOptions :: Parser Command
updateOptions = Update <$> argument str (metavar "PEER" <> help "Peer to update with")

updateAllOptions :: Parser Command
updateAllOptions = pure UpdateAll

uuxOptions :: Parser Command
uuxOptions = Uux <$> argument auto (metavar "PEER-ID" <> help "Peer on which to execute the comand")
                 <*> argument str (metavar "cmd" <> help "The command to execute remotely, e.g., /usr/sbin/sendmail")
                 <*> many (argument str (metavar "arg" <> help "argument for cmd"))

dropPeerOptions :: Parser Command
dropPeerOptions = DropPeer <$> argument auto (metavar "PEER-ID" <> help "Peer to drop from our network")

withOptions :: (Options -> RIO env a) -> RIO env a
withOptions = (liftIO (execParser optsWithInfo) >>=)
