module Data
       ( MailFlags
       , fromMailFlags
       , emptyFlags
       , parseMailFlags
       , serializeMailFlags
       , FSName(..)
       , fromFSName
       , Id(..)
       , nullId
       , GlobalName
       , createGlobalName
       , fromGlobalName
       , Mail(..)
       , OnDiskMail(..)
       , toOnDiskMail
       , InitialSyncMail(..)
       , InitialSyncState(..)
       , InitialMessage(..)
       , MailBox(..)
       , createMailBox
       , State
       , MailContent
       , renderMailContent
       , mailContentFromByteString
       , Operation(..)
       , Patch(..)
       , ChangePayload(..)
       , UUX (..)
       , Change(..)
       , History
       , PeerState
       , PeerId
       , UpstreamPeers
       , GlobalState(..)
       , emptyGlobalState
       , globalStateFileName
       , DownstreamPeer(..)
       )
where

import RIO
import RIO.HashMap as H
import RIO.Set

import Data.Binary
import qualified Data.ByteString.Char8 as B
import Data.UUID (UUID)

import IOClasses

-- | Type of flags a mail can have in a mail box
newtype MailFlags = MailFlags { fromMailFlags' :: ByteString }
    deriving (Show, Read, Eq, Ord, Hashable, Generic)
instance Binary MailFlags

fromMailFlags :: MailFlags -> String
fromMailFlags = B.unpack . fromMailFlags'

emptyFlags :: MailFlags
emptyFlags = MailFlags { fromMailFlags' = "" }

parseMailFlags :: String -> MailFlags
parseMailFlags = MailFlags . fromString

serializeMailFlags :: OnDiskMail -> String
serializeMailFlags = fromMailFlags . odFlags

-- | Type of names acceptable in the file system
newtype FSName = FSName { fromFSName' :: ByteString }
    deriving (Show, Read, Eq, Ord, Hashable, Generic)
instance Binary FSName

fromFSName :: FSName -> String
fromFSName = B.unpack . fromFSName'

newtype Id = Id { fromId :: Integer } deriving (Show, Read, Eq, Ord, Generic)
instance Binary Id

nullId :: Id
nullId = Id { fromId = 0 }

-- | Type of global names; for the time being only Strings,
-- but something more structured might be desirable (we'll see).
newtype GlobalName = GlobalName { fromGlobalName' :: ByteString}
    deriving (Show, Read, Eq, Ord, Hashable, Generic)
instance Binary GlobalName

fromGlobalName :: GlobalName -> String
fromGlobalName = B.unpack . fromGlobalName'

-- | Generate a global name from a UUID and an internal id
createGlobalName :: UUID -> Id -> GlobalName
createGlobalName uuid n
    = GlobalName { fromGlobalName' = fromString (
                       (show uuid)
                       ++ "-"
                       ++ show (fromId n)) }

data Mail = Mail { mBase   :: FSName
                 , mglobalName :: GlobalName
                 , mFlags      :: MailFlags
                 } deriving (Show, Read, Eq, Ord, Generic)

instance Binary Mail

data OnDiskMail = OnDiskMail
  { odBase :: FSName
  , odFlags    :: MailFlags
  } deriving (Show, Read, Eq, Ord)

toOnDiskMail :: Mail -> OnDiskMail
toOnDiskMail mail = OnDiskMail { odBase = mBase mail
                               , odFlags = mFlags mail
                               }

data InitialSyncMail = InitialSyncMail 
                       { iMailBox :: MailBox
                       , iGlobalName :: GlobalName
                       , iFlags :: MailFlags
                       , iBody :: ByteString
                       } deriving (Show, Read, Eq, Generic)

instance Binary InitialSyncMail 

data InitialSyncState = InitialSyncState { iMyId :: PeerId
                                         , iMyName :: String
                                         , iYouKnowUntil :: PeerState
                                         }
  deriving (Show, Read, Eq, Generic)

instance Binary InitialSyncState

data InitialMessage = InitialMail InitialSyncMail 
                    | InitialDoneSync InitialSyncState
  deriving (Show, Read, Eq, Generic)

instance Binary InitialMessage

newtype MailBox = MailBox { fromMailbox :: FSName }
                deriving (Show, Read, Eq, Ord, Hashable, Generic)
instance Binary MailBox

createMailBox :: String -> MailBox
createMailBox name = MailBox { fromMailbox = FSName { fromFSName' = fromString name } }

-- | State of all mail boxes as we saw them the last time
type State = HashMap MailBox (Set Mail)

-- | Type of the contents of a mail
newtype MailContent = MailContent { fromMailContent :: ByteString }
    deriving (Show, Read, Eq, Ord, Generic)
instance Binary MailContent

renderMailContent :: HasMailSyncConfig env => MailContent -> RIO env ByteString
renderMailContent = return . fromMailContent

mailContentFromByteString :: ByteString -> MailContent
mailContentFromByteString bs = MailContent { fromMailContent = bs }

data Operation = Add MailContent MailFlags
               | Delete
               | Flags MailFlags
               deriving (Show, Read, Eq, Generic)
instance Binary Operation

instance Ord Operation where
  Add content flags <= Add content' flags'
     = content < content' || (content == content' && flags <= flags')
  Add _ _     <=  _           = True
  Flags _     <= Add _ _      = False
  Flags flags <= Flags flags' = flags <= flags'
  Flags _     <= _            = True
  Delete      <= Delete       = True
  Delete      <= _            = False

data Patch = Patch { mailbox :: MailBox
                   , globalName :: GlobalName
                   , operation  :: Operation
                   }
             deriving (Show, Read, Eq, Ord, Generic)
instance Binary Patch

data UUX = UUX { uuxPeer ::  PeerId
               , uuxStdin :: ByteString
               , uuxCmd :: (String, [String])
               }
           deriving (Show, Read, Eq, Ord, Generic)
instance Binary UUX

data ChangePayload = ChangePatch Patch | ChangeUUX UUX
           deriving (Show, Read, Eq, Ord, Generic)
instance Binary ChangePayload

data Change = Change { changeId :: Id
                     , changePatch :: ChangePayload
                     , changeSource :: UUID
                     } deriving (Show, Read, Eq, Generic)
instance Binary Change

instance Ord Change where
    Change ident patch uuid <= Change ident' patch' uuid'
        = ident < ident' ||
          (ident == ident' && ( patch < patch' ||
                                patch == patch' && uuid <= uuid'))


type History = Set Change

type PeerState = Id

type PeerId = UUID

type UpstreamPeers = HashMap PeerId PeerState

instance (Eq k, Hashable k, Binary k, Binary v) => Binary (HashMap k v) where
  get = H.fromList <$> get
  put = put . H.toList

data GlobalState = GlobalState { state :: State
                               , history :: History
                               , nextId :: Id
                               , upstreamPeers :: UpstreamPeers
                               , myUUID :: UUID
                               , myName :: String
                               }
                 deriving (Show, Read, Generic)

instance Binary GlobalState

emptyGlobalState :: String -> UUID -> GlobalState
emptyGlobalState name uuid
    = GlobalState { state = H.empty
                  , history = RIO.Set.empty
                  , nextId = Id { fromId = 1}
                  , upstreamPeers = H.empty
                  , myUUID = uuid
                  , myName = name
                  }

globalStateFileName :: FilePath
globalStateFileName = "state.global"

data DownstreamPeer = DownstreamPeer PeerId String PeerState
                    deriving (Show, Read)
