module IOData
where

import RIO
import qualified RIO.ByteString.Lazy as LB
import RIO.File (writeBinaryFileDurableAtomic, withBinaryFileDurableAtomic)
import RIO.FilePath ((</>))
import RIO.Directory (getHomeDirectory, createDirectoryIfMissing)
import RIO.Process (HasProcessContext(..), ProcessContext
                   )
import qualified RIO.Text as T

import Data.Binary (decodeFileOrFail, encode)
import Network.BSD (getHostName)
import System.Posix.Process (getProcessID)
import System.Posix.Time (epochTime)

import IOClasses
import Utils
import Cli
import ExitCodes
import Data (GlobalState, globalStateFileName, nullId, PeerId, PeerState)
import Data.UUID
import Lock.Sync
import Match.Predicate
import Peer

data MailSync = MailSync
              { msLogFunc :: !LogFunc
              , msProcessContext :: !ProcessContext
              , msMailRoot :: FilePath
              , msDeliveryCount :: TVar Integer
              , msDeliveryPrefix :: Integer -> String
              , msSyncDir :: FilePath
              , msOptions :: Options
              , msPredicates :: [Predicate]
              , msAcceptableCmds :: [String]
              , msPeers :: [Peer]
              , msMyCmd :: (String, [String])
              }

instance HasLogFunc MailSync where
  logFuncL = lens msLogFunc (\x y -> x { msLogFunc = y })
instance HasProcessContext MailSync where
  processContextL = lens msProcessContext (\x y -> x { msProcessContext = y})

instance HasMailRoot MailSync where
  mailDirRootL = lens msMailRoot (\x y -> x { msMailRoot = y})

instance HasMailDeliveryInfo MailSync where
  mailDeliveryCountL = lens msDeliveryCount (\x y -> x { msDeliveryCount = y})
  mailDeliveryPrefixL = lens msDeliveryPrefix (\x y -> x { msDeliveryPrefix = y})

instance HasMailSyncConfig MailSync where
  mailSyncDirL = lens msSyncDir (\x y -> x { msSyncDir = y})
  mailSyncPredicatesL = lens msPredicates (\x y -> x { msPredicates = y})
  mailSyncAcceptableCmdsL = lens msAcceptableCmds (\x y -> x { msAcceptableCmds = y})
  mailSyncPeersL = lens msPeers (\x y -> x { msPeers = y})
  mailSyncMyCmdL = lens msMyCmd (\x y -> x { msMyCmd = y})

instance HasOptions MailSync where
  cliOptionsL = lens msOptions (\x y -> x { msOptions = y})

run :: RIO MailSync a -> IO ()
run inner = runSimpleApp . void $ runMailSync inner

runMailSync :: (HasProcessContext env, HasLogFunc env)
             => RIO MailSync a -> RIO env a
runMailSync inner = do
  withOptions $ \options -> do
    config <- readConfigFile (cfgFile options) $ optInitialLog options
    createDirectoryIfMissing True $ syncDir config
    let logFile = syncDir config </> "log"
    withBinaryFile logFile AppendMode $ \logHandle -> do
      logOptions <- logOptionsHandle logHandle True
      withLogFunc logOptions $ \logFunc -> do
        processContext <- view processContextL
        initialDeliveryCount <- newTVarIO 0
        hostName <- liftIO getHostName
        time <- show <$> liftIO epochTime
        pid <- show <$> liftIO getProcessID
        let app = MailSync
              { msLogFunc = logFunc
              , msProcessContext = processContext
              , msMailRoot = mailRoot config
              , msDeliveryCount = initialDeliveryCount
              , msDeliveryPrefix = \count -> time ++ "." ++ pid ++ "_" ++ show count ++ "." ++ hostName
              , msSyncDir = syncDir config
              , msOptions = options
              , msPredicates = predicates config
              , msAcceptableCmds = acceptableCmds config
              , msPeers = myPeers config
              , msMyCmd = myCmd config
              } in runRIO app inner

sampleInner :: HasLogFunc env => RIO env ()
sampleInner = logInfo "Test"

readText :: Read a => ByteString -> Maybe a
readText = (rightToMaybe . decodeUtf8') >=> (readMaybe . T.unpack)

showText :: Show a => a -> ByteString
showText = encodeUtf8 . T.pack . show

withGlobalStateLocked ::
  (HasMailSyncConfig env) =>
  (GlobalState -> RIO env a) -> RIO env a
withGlobalStateLocked inner = withMasterLock Exclusive $ do
  state <- slurpGlobalState
  inner state

slurpGlobalState ::
  (HasMailSyncConfig env) => RIO env GlobalState
slurpGlobalState =
  tryAny getGlobalState' >>= either (\e -> do
    printErrorMessage $ "Failed to read global state: " ++ show e
    exitWith $ ExitFailure exitFailedToReadGlobalState) return


getGlobalState' :: (HasMailSyncConfig env) => RIO env GlobalState
getGlobalState' = do
  syncDir <- view mailSyncDirL
  mState <- liftIO $ rightToMaybe <$> decodeFileOrFail (syncDir </> globalStateFileName)
  maybe (throwIO InvalidGlobalStateException) return mState

readConfigFile :: HasLogFunc env => FilePath -> Bool -> RIO env Config
readConfigFile configFile' shouldLog = do
  configFile <- if take 2 configFile' == "~/"
                   then do
                        home <- getHomeDirectory
                        return $ home </> drop 2 configFile'
                   else return configFile'
  when shouldLog . logInfo . display $ "Loading config file: " `T.append` T.pack configFile
  (mConfig :: Maybe Config) <- readText <$> readFileBinary configFile
  when shouldLog . logInfo . display $ "Loaded Config: " `T.append` ((T.pack . show) mConfig)
  maybe (throwIO InvalidConfigFileException) (return) mConfig

writeConfigFile :: HasOptions env
                => Config -> RIO env ()
writeConfigFile config = do
  configFile' <- cfgFile <$> view cliOptionsL
  configFile <- if take 2 configFile' == "~/"
                   then do
                        home <- getHomeDirectory
                        return $ home </> drop 2 configFile'
                   else return configFile'
  writeFileBinary configFile . encodeUtf8 . T.pack $ show config

addPeer' :: (String, UUID) -> Config -> Config
addPeer' (name, uuid) cfg = cfg { myPeers = myPeers cfg ++ [newPeer] }
  where newPeer = Peer { pName = name, pUUID = uuid, pPath = Nothing }

addPeer :: (HasOptions env, HasLogFunc env)
        => (String, UUID) -> RIO env ()
addPeer peer = do
 configFile <- cfgFile <$> view cliOptionsL
 (addPeer' peer <$> readConfigFile configFile True) >>= writeConfigFile

data Config = Config
  { mailRoot :: String
  , syncDir :: String
  , predicates :: [Predicate]
  , acceptableCmds :: [String]
  , myPeers :: [Peer]
  , myCmd :: (String, [String])
  } deriving (Show, Read, Eq)

data MailSyncException = InvalidConfigFileException
                       | InvalidGlobalStateException
                       | InvalidPeerStateException
  deriving (Show, Read, Eq)

instance Exception MailSyncException

writeGlobalState :: HasMailSyncConfig env => GlobalState -> RIO env GlobalState
writeGlobalState gs = do
  syncDir <- view mailSyncDirL
  withBinaryFileDurableAtomic (syncDir </> globalStateFileName) WriteMode $ \h ->
    LB.hPut h $ encode gs
  return gs

writeGlobalState_ :: HasMailSyncConfig env => GlobalState -> RIO env ()
writeGlobalState_ = void . writeGlobalState

writePeerState :: HasMailSyncConfig env => PeerId -> PeerState -> RIO env ()
writePeerState peer state = do
 syncDir <- view mailSyncDirL
 let peersDir = syncDir </> "peers"
 createDirectoryIfMissing True peersDir
 writeBinaryFileDurableAtomic (peersDir </> (show peer)) $ showText state

readPeerState :: HasMailSyncConfig env => PeerId -> RIO env PeerState
readPeerState peer = do
  syncDir <- view mailSyncDirL
  (mState :: Maybe PeerState) <- exceptNothing $
    readText <$> (readFileBinary $ syncDir </> "peers" </> show peer)
  return $ fromMaybe nullId mState
