module InitialSync ( initialSyncForPeer
                   , initialSyncFrom
)
where

import RIO
import RIO.Directory (doesDirectoryExist)
import Data.Binary (encode, decodeOrFail)
import Data.Binary.Get (ByteOffset)

import qualified RIO.ByteString.Lazy as LB
import qualified RIO.ByteString as BS
import qualified RIO.HashMap as H
import qualified RIO.Set as S
import qualified RIO.Text as T

import qualified ForeignUtils as FF

import Data
import IOClasses
import IOData ( withGlobalStateLocked
              , writeGlobalState
              , writeGlobalState_
              , writePeerState
              , addPeer
              )
import ExitCodes
import LocalUpdates (getMails, removeLostEntries)
import Maildir (findMail, newMailQD)
import Peer (pName)
import RemoteUpdates (acceptPeer, anyPeer)
import Utils (maybeReadBinaryFile, (>>=?), printErrorMessage)

initialSyncForPeer :: ( HasMailSyncConfig env
                      , HasMailRoot env
                      , HasLogFunc env
                      , HasOptions env
                      )
                      => PeerId -> String -> RIO env ()
initialSyncForPeer peer name = withGlobalStateLocked $ \origGs -> do
  ensureInitialSyncForPermitted origGs peer name
  let mailboxes = H.keys $ state origGs
  origGs' <- foldM `flip` origGs `flip` mailboxes $ \st mb -> do
    let mails = fromMaybe S.empty . H.lookup mb $ state st
    foldM `flip` st `flip` mails $ \ gs mail -> gs `seq` do
      mbContent <- findMail mb (toOnDiskMail mail) >>=? maybeReadBinaryFile
      maybe (writeGlobalState . removeLostEntries gs mb . S.singleton $ mglobalName mail)
         `flip` mbContent
         $ \ content -> do
               let toSend = InitialMail $ InitialSyncMail { iMailBox = mb
                                                          , iGlobalName = mglobalName mail
                                                          , iFlags = mFlags mail
                                                          , iBody = content
                                                          }
               BS.hPut stdout . toStrictBytes $ encode toSend
               return gs
  let peerState = Id { fromId = (fromId $ nextId origGs') -1 }
  let doneMessage = InitialDoneSync 
                    $ InitialSyncState 
                      { iMyId = myUUID origGs'
                      , iMyName = myName origGs'
                      , iYouKnowUntil = peerState
                      }
  let gs' = acceptPeer origGs' peer
  writeGlobalState_ gs'
  writePeerState peer peerState
  addPeer (name, peer)
  BS.hPut stdout . toStrictBytes $ encode doneMessage

ensureInitialSyncForPermitted :: ( HasLogFunc env 
                                 , HasMailSyncConfig env 
                                 )
                              => GlobalState -> PeerId -> String -> RIO env ()
ensureInitialSyncForPermitted gs peer name = do
  when (peer `H.member` (upstreamPeers gs)) $ do
    printErrorMessage $ "Initial sync for not permitted: peer " ++ show peer ++ " already known"
    exitWith $ ExitFailure exitPeerKnown
  peers <- view mailSyncPeersL
  when (name `elem` (map pName peers)) $ do
    let msg = "WARNING: a peer with nick name " ++ name ++ " is already known; continuing anyway"
    logWarn . display $ T.pack msg
    printErrorMessage msg


initialParseFailure :: (LB.ByteString, ByteOffset, String) -> RIO env ()
initialParseFailure (_, _, e) = do
  printErrorMessage $ "Failed to parse initial message: " ++ show e
  exitWith $ ExitFailure exitFailedInitialParse

initialSyncFrom :: ( HasMailSyncConfig env
                   , HasLogFunc env
                   , HasMailRoot env
                   , HasMailDeliveryInfo env
                   , HasOptions env
                   )
                => RIO env ()
initialSyncFrom = withGlobalStateLocked $ \gs -> do
  ensureInitialSyncFromPermitted gs
  LB.hGetContents stdin >>= initialSyncFrom' gs

ensureInitialSyncFromPermitted :: ( HasMailRoot env)
                               => GlobalState -> RIO env ()
ensureInitialSyncFromPermitted gs = do
  when (anyPeer gs) $ do
     printErrorMessage "Peers present, refusing to receive initial Sync"
     exitWith $ ExitFailure exitAlreadyPeers
  root <- view mailDirRootL
  whenM (doesDirectoryExist root) $ do
    printErrorMessage $ "Mail root: " ++ root ++ " present, refusing to recieve initial Sync"
    exitWith $ ExitFailure exitMailRootPresent

initialSyncFrom' :: ( HasMailSyncConfig env
                    , HasLogFunc env
                    , HasMailRoot env
                    , HasMailDeliveryInfo env
                    , HasOptions env
                    )
                 => GlobalState -> LB.ByteString -> RIO env ()
initialSyncFrom' gs input = do
  let parsed = decodeOrFail input
  either initialParseFailure `flip` parsed $ \(input', _, msg) -> do
    case msg of 
      InitialDoneSync s  -> do
        FF.sync
        writeGlobalState_ (gs { upstreamPeers = H.singleton (iMyId s) (iYouKnowUntil s) })
        addPeer (iMyName s, iMyId s)
      InitialMail m -> do
        let mb = iMailBox m
        baseName <- newMailQD mb (iBody m) (iFlags m)
        let gn = iGlobalName m
        let flags = iFlags m
        let mail = Mail { mBase = baseName
                        , mglobalName = gn
                        , mFlags = flags
                        }
        let newset = S.insert mail $ getMails mb gs
        let state' = H.insert mb newset $ state gs
        let gs' = gs { state = state' }
        gn `seq` flags `seq` mail `seq` newset `seq` state' `seq` gs' `seq` input' `seq` initialSyncFrom' gs' input'
