module LocalUpdates ( addNewMail
                    , addUux
                    , getMails
                    , getNextId
                    , new
                    , purge
                    , removeLostEntries
                    , update
                    ) where

import Data
import RIO
import qualified RIO.HashMap as H
import qualified RIO.Set as S

-- | From global state, get the known mails in the given mailbox
getMails :: MailBox -> GlobalState -> Set Mail
getMails mailbox = fromMaybe S.empty . H.lookup mailbox . state

-- | Given the currently known mails and the entries found on disk,
-- compute the file names that were newly found on disk
newEntries :: Set Mail -> Set OnDiskMail -> Set FSName
newEntries known onDisk
    = (S.map odBase onDisk) S.\\ (S.map mBase known)

-- | Given the currently known mails and the entries found on disk,
-- compute the list of global names that have been removed.
lostEntries :: Set Mail -> Set OnDiskMail -> Set GlobalName
lostEntries known onDisk
    = let remainingFiles = S.map odBase onDisk
          lost = S.filter
                 (flip S.notMember remainingFiles . mBase)
                 known
      in S.map mglobalName lost

-- | Given an Id and a count, generate that many new Ids and also
-- return the next free Id.
getNewIds :: Id -> Integer -> ([Id], Id)
getNewIds ident count =
    let n = fromId ident
    in (map Id [n .. (n + count - 1)], Id $ n + count)

-- | Given an Id, generate the next one
getNextId :: Id -> Id
getNextId ident = Id $ (fromId ident) + 1

-- | Given a global state, a mailbox and a list of global names,
-- that were lost on Disk, update the global state accordingly.
-- This includes generating new patches for what was observed.
removeLostEntries :: GlobalState -> MailBox -> Set GlobalName -> GlobalState
removeLostEntries st mb lost =
    let lostList = S.toAscList lost
        (newIds, newNextId) = getNewIds (nextId st) . fromIntegral
                              $ length lostList
        newPatches = map (\ gn -> Patch { mailbox = mb
                                        , operation = Delete
                                        , globalName = gn
                                        }) lostList
        newChanges = zipWith
                     (\ ident patch -> Change { changeId = ident
                                              , changePatch = ChangePatch patch
                                              , changeSource = myUUID st
                                              })
                     newIds newPatches
        newHistory = S.union (history st) $ S.fromList newChanges
        newMails = S.filter (flip S.notMember lost .mglobalName)
                   $ getMails mb st
        newState = H.insert mb newMails $ state st
    in st { state = newState
          , nextId = newNextId
          , history = newHistory
          }

-- | Compare a mail the on-disk representation and update flags as needed.
updateFlagsMail :: (Mail -> Maybe MailFlags)
                -> MailBox -> Mail -> (Mail, [Patch])
updateFlagsMail lookupFlags mb m =
    case lookupFlags m of
         Nothing -> (m, [])
         Just flags -> if mFlags m == flags
                       then (m, [])
                       else ( m {mFlags = flags }
                            , [ Patch { mailbox = mb
                                      , globalName = mglobalName m
                                      , operation = Flags flags
                                      } ])

-- | Compare a mailbox  with its on-dist representation
updateFlags :: MailBox -> Set OnDiskMail -> Set Mail -> (Set Mail, [Patch])
updateFlags mb ondisk ms =
    let flagsMap = H.fromList
                   . map (\m -> (odBase m, odFlags m))
                   $ S.toAscList ondisk
        lookupFlags m = H.lookup (mBase m) flagsMap
        updates = map (updateFlagsMail lookupFlags mb)
                  $ S.toAscList ms
    in (S.fromList . map fst $ updates, concat . map snd $ updates)

-- | Given a global state, a mailbox and its on-disk values, return an updated
-- state and set of files that still have to be processed as newly discovered.
update :: GlobalState
       -> MailBox
       -> Set OnDiskMail
       -> (GlobalState, Set OnDiskMail)
update gs mb ondisk =
    let mails = getMails mb gs
        lost = lostEntries mails ondisk
        (mails', flagPatches) = updateFlags mb ondisk mails
        (newIds, newNextId) = getNewIds (nextId gs) . fromIntegral
                              $ length flagPatches

        flagChanges = zipWith
                      (\ ident patch -> Change { changeId = ident
                                               , changePatch = ChangePatch patch
                                               , changeSource = myUUID gs
                                               })
                      newIds flagPatches
        history' = S.union (history gs) $ S.fromList flagChanges
        gs' = gs { nextId = newNextId
                 , history = history'
                 , state = H.insert mb mails' $ state gs
                 }
        newOnDisk = newEntries mails ondisk
        in ( removeLostEntries gs' mb lost
           , S.filter (flip S.member newOnDisk . odBase) ondisk
           )

-- | Given a global state, a mailbox, and a mail newly discovered there, update
-- the global state accordingly.
addNewMail :: GlobalState
           -> MailBox
           -> (OnDiskMail, MailContent)
           -> GlobalState
addNewMail gs mb (ondisk, content) =
    let patchId = nextId gs
        gname = createGlobalName (myUUID gs) patchId
        flags = odFlags ondisk
        mail = Mail { mBase = odBase ondisk
                    , mglobalName = gname
                    , mFlags = flags
                    }
        mails' = S.insert mail $ getMails mb gs
        patch = Patch { mailbox = mb
                      , globalName = gname
                      , operation = Add content flags
                      }
    in gs { nextId = getNextId patchId
          , state = H.insert mb mails' $ state gs
          , history = S.insert
                      (Change { changeId = patchId
                              , changePatch = ChangePatch patch
                              , changeSource = myUUID gs
                              })
                      $ history gs
          }


-- | Accept a uux request in the global state
addUux :: GlobalState
       -> PeerId
       -> (String, [String])
       -> ByteString
       -> GlobalState
addUux gs peer cmd input =
    let uux = UUX { uuxPeer = peer
                  , uuxStdin = input
                  , uuxCmd = cmd
                  }
        patchId = nextId gs
    in gs { nextId = getNextId patchId
          , history = S.insert
                      (Change { changeId = patchId
                              , changePatch = ChangeUUX uux
                              , changeSource = myUUID gs
                              })
                      $ history gs
          }

-- | Purge entries from history that are known to every peer.
purge :: [PeerState] -> GlobalState -> GlobalState
purge peers gs =
    let cutoff = foldr min (nextId gs) peers
    in gs { history = S.filter ((cutoff <=) . changeId) $ history gs }

insertOnDiskMail :: MailBox -> OnDiskMail -> GlobalState -> GlobalState
insertOnDiskMail mb odm gs =
    let patchId = nextId gs
        gname = createGlobalName (myUUID gs) patchId
        flags = odFlags odm
        mail = Mail { mBase = odBase odm
                    , mglobalName = gname
                    , mFlags = flags
                    }
        mails' = S.insert mail $ getMails mb gs
    in gs { nextId = getNextId patchId
          , state = H.insert mb mails' $ state gs
          }

-- | Given a global state, a mailbox an its on-disk values, return the
-- state with the mailboxes state filled from scratch without recording
-- history
new :: GlobalState
    -> MailBox
    -> Set OnDiskMail
    -> GlobalState
new gs mb ondisk = foldr (insertOnDiskMail mb) gs ondisk
