module Maildir 
  ( getMails
  , findMail
  , changeFlags
  , newMail, newMailQD
  , deleteMail
  , getMailboxes
  )
where

import RIO
import RIO.Directory (listDirectory
                     , renameFile
                     , removeFile
                     , doesFileExist
                     , doesDirectoryExist
                     , createDirectoryIfMissing
                     )
import RIO.File (writeBinaryFileDurableAtomic)
import RIO.FilePath  ((</>), combine)
import RIO.List (find, isPrefixOf)
import qualified RIO.Set as S

import Text.Parsec as P

import Data
import IOClasses
import Utils

mailNameParser :: P.Parsec String st String
mailNameParser = P.many $ P.noneOf [':']

mailFlagsParser :: P.Parsec String st (Maybe String)
mailFlagsParser = optionMaybe $ (P.string ":2,") *> P.many anyChar

mailParser :: P.Parsec String st (String, Maybe String)
mailParser = (,) <$> mailNameParser <*> mailFlagsParser

constructOnDiskMail :: (String, Maybe String) -> OnDiskMail
constructOnDiskMail (name, flags) = 
  OnDiskMail { odBase = fsName
             , odFlags = flags'
             }
  where fsName = FSName { fromFSName' = fromString name }
        flags' = maybe emptyFlags parseMailFlags flags

parseOnDiskMail :: FilePath -> Maybe OnDiskMail
parseOnDiskMail fp = constructOnDiskMail <$> (rightToMaybe $ P.parse mailParser "" fp)

getMails :: HasMailRoot env => MailBox -> RIO env (Maybe (Set OnDiskMail))
getMails wrapMaildir = do
  let maildir = fromFSName $ fromMailbox  wrapMaildir
  root <- view mailDirRootL
  curr <- listDirectory (root </> maildir </> "cur")
  new  <- listDirectory (root </> maildir </> "new")
  let mails = curr ++ new
  return $ S.fromList <$> mapM parseOnDiskMail mails

constructMailFile :: MailBox -> OnDiskMail -> FilePath
constructMailFile mb odm =
  mailBoxPath </> "cur" </> mailPath
  where mailbase = fromFSName $ odBase odm
        mailPath = mailbase ++ ":2," ++ serializeMailFlags odm
        mailBoxPath = fromFSName $ fromMailbox mb

constructNewMailFile :: MailBox -> OnDiskMail -> FilePath
constructNewMailFile mb odm =
  mailBoxPath </> newcur </> mailPath
  where mailbase = fromFSName $ odBase odm
        mailBoxPath = fromFSName $ fromMailbox mb
        (mailPath, newcur) = if (serializeMailFlags odm == "")
                               then (mailbase, "new")
                               else (mailbase ++ ":2," ++ serializeMailFlags odm, "cur")


findMail :: (HasMailRoot env) => MailBox -> OnDiskMail -> RIO env (Maybe FilePath)
findMail mb odm = do
 root <- view mailDirRootL
 let prefix = root </> (fromFSName $ fromMailbox mb)
 atNew <- doesFileExist $ prefix </> "new" </> base
 if atNew then return . return $ prefix </> "new" </> base else do
   let newWithFlags = prefix </> "new" </> base ++ ":2," ++  flags
   atNewWithFlags <- doesFileExist newWithFlags
   if atNewWithFlags then return $ return newWithFlags else do
     let curWithFlags = prefix </> "cur" </> base ++ ":2," ++ flags
     atCurWithFlags <- doesFileExist curWithFlags
     if atCurWithFlags then return $ return curWithFlags else do
       filesInCurr <- listDirectory $ prefix </> "cur"
       let found = find (isPrefixOf $ base ++ ":") filesInCurr
       found `seq` (return $ (combine $ prefix </> "cur") <$> found)
 where base = fromFSName $ odBase odm
       flags = serializeMailFlags odm

changeFlags :: HasMailRoot env => MailBox -> MailFlags -> OnDiskMail -> RIO env (Maybe ())
changeFlags mb newFlags odm = do
  root <- view mailDirRootL
  let new = constructMailFile mb odm { odFlags = newFlags }
  findMail mb odm >>=?? (renameFile `flip` (root </> new))

getCurrDeliveryCountAndIncrease :: TVar Integer -> STM Integer
getCurrDeliveryCountAndIncrease t = do
  curr <- readTVar t
  writeTVar t (curr + 1)
  return curr

newMail' :: (HasMailRoot env, HasMailDeliveryInfo env)
         => (FilePath -> ByteString -> RIO env ()) 
         -> (a -> RIO env ByteString)
         -> MailBox -> a -> MailFlags -> RIO env FSName
newMail' doPersist renderContent mb content flags = do
  root <- view mailDirRootL
  countSTM <- view mailDeliveryCountL 
  mailDeliveryPrefix <- view mailDeliveryPrefixL 
  currCount <- atomically $ getCurrDeliveryCountAndIncrease countSTM
  let base = FSName . fromString $ mailDeliveryPrefix currCount
  let fp = constructNewMailFile mb $ OnDiskMail base flags
  createMailBoxIfDoesNotExist mb
  mContent <- renderContent content
  doPersist (root </> fp) mContent
  return base


newMail :: (HasMailRoot env, HasMailDeliveryInfo env, HasMailSyncConfig env)
        => MailBox -> MailContent -> MailFlags -> RIO env FSName
newMail = newMail' writeBinaryFileDurableAtomic renderMailContent 
  -- TODO: Write Durable to tmp, then move to cur
  -- TODO: Needs isEmptyFlags which tells we want to deliver to new

newMailQD :: (HasMailRoot env, HasMailDeliveryInfo env)
          => MailBox -> ByteString -> MailFlags -> RIO env FSName
newMailQD = newMail' writeFileBinary return 
  -- TODO: maybe to do the rename (but not the sync) dance

deleteMail :: HasMailRoot env => MailBox -> OnDiskMail -> RIO env (Maybe ())
deleteMail mb odm = findMail mb odm >>=?? removeFile

getMailboxes :: HasMailRoot env => RIO env (Set MailBox)
getMailboxes = do
  root <- view mailDirRootL
  allInRoot <- listDirectory root
  S.fromList <$> map (MailBox . FSName . fromString) <$> filterM (\x -> doesDirectoryExist (root </> x)) allInRoot

createMailBoxIfDoesNotExist :: HasMailRoot env => MailBox -> RIO env ()
createMailBoxIfDoesNotExist mb = do
  root <- view mailDirRootL
  mapM_ (createDirectoryIfMissing True
         . ((root </> (fromFSName $ fromMailbox mb)) </>)) ["tmp", "new", "cur"]
