module RemoteUpdates ( myBase
                     , acceptIndexOnly
                     , acceptNonApply
                     , acceptFlagChange
                     , acceptMailRemoval
                     , acceptNewMail
                     , acceptPeer
                     , anyPeer
                     , dropPeer
                     , peerKnown
                     ) where

import Data
import LocalUpdates (getMails, getNextId)
import RIO
import qualified RIO.HashMap as H
import qualified RIO.Set as S

-- | Given a global state, checks whether we know any peer
anyPeer :: GlobalState -> Bool
anyPeer gs = not . H.null $ upstreamPeers gs

-- | Given a global state and a peer identifier, indicate whether the
-- peer is already known as upstream peer for us.
peerKnown :: GlobalState -> PeerId -> Bool
peerKnown gs peer = H.member peer $ upstreamPeers gs

-- | Givel a global state and a peer identifier, accept the peer as
-- new upstream peer (from which we don't know anything), if not already
-- known.
acceptPeer :: GlobalState -> PeerId -> GlobalState
acceptPeer gs peer =
  if peerKnown gs peer
     then gs
     else gs { upstreamPeers = H.insert peer nullId $ upstreamPeers gs }

-- | Given the UUID of a peer, drop it in a given global configuration
dropPeer :: GlobalState -> PeerId -> GlobalState
dropPeer gs peer = gs { upstreamPeers = H.delete peer $ upstreamPeers gs }

-- | Given a mailbox and a global name, determine the local base
-- name for that mail.
myBase :: GlobalState -> MailBox -> GlobalName -> Maybe OnDiskMail
myBase gs mb gname =
    fmap toOnDiskMail
    . S.lookupMin
    . S.filter ((==) gname . mglobalName)
    $ getMails mb gs


acceptIndexOnly :: PeerId
                -> PeerState
                -> GlobalState
                -> GlobalState
acceptIndexOnly peer peerState gs =
  gs { upstreamPeers = H.insert peer peerState $ upstreamPeers gs }


acceptNonApply :: PeerId
               -> PeerState
               -> GlobalState
               -> Patch
               -> GlobalState
acceptNonApply peer peerState gs patch =
    let patchId = nextId gs
        change = Change { changeId = patchId
                        , changePatch = ChangePatch patch
                        , changeSource = peer
                        }
    in gs { nextId = getNextId patchId
          , history = S.insert change $ history gs
          , upstreamPeers = H.insert peer peerState $ upstreamPeers gs
          } 

acceptPatch :: PeerId
            -> PeerState
            -> GlobalState
            -> MailBox
            -> GlobalName
            -> Operation
            -> (Set Mail -> Set Mail)
            -> GlobalState
acceptPatch peer peerState gs mb gname operation mbTransform =
    let patchId = nextId gs
        change = Change { changeId = patchId
                        , changePatch = ChangePatch
                                          $ Patch { mailbox = mb
                                                  , globalName = gname
                                                  , operation = operation
                                                  }
                         , changeSource = peer
                         }
    in gs { nextId = getNextId patchId
          , state = H.insert mb
                    (mbTransform $ getMails mb gs)
                    $ state gs
          , history = S.insert change $ history gs
          , upstreamPeers = H.insert peer peerState $ upstreamPeers gs
          } 

-- | Record in the global state that a flag-change request by a given
-- peer was succesfully carried out.
acceptFlagChange :: PeerId
                 -> PeerState
                 -> GlobalState
                 -> MailBox
                 -> GlobalName
                 -> MailFlags
                 -> GlobalState
acceptFlagChange peer peerState gs mb gname flags =
    acceptPatch peer peerState gs mb gname (Flags flags) $ \mbState ->
    let (affectedMail, otherMails)
            = S.partition ((==) gname . mglobalName) $ mbState
        affectedMail' = S.map (\ m -> m { mFlags = flags }) affectedMail
    in S.union affectedMail' otherMails

-- | Record in the global state that a mail-removal request by a given peer
-- was successfully carried out.
acceptMailRemoval :: PeerId
                  -> PeerState
                  -> GlobalState
                  -> MailBox
                  -> GlobalName
                  -> GlobalState
acceptMailRemoval peer peerState gs mb gname =
    acceptPatch peer peerState gs mb gname Delete $ S.filter ((/=) gname . mglobalName)

-- | Record in the global state that a new-mail request by a given peer
-- was successfully carried out.
acceptNewMail :: PeerId
              -> PeerState
              -> GlobalState
              -> MailBox
              -> GlobalName
              -> FSName
              -> MailContent
              -> MailFlags
              -> GlobalState
acceptNewMail peer peerState gs mb gname lname content flags =
    let operation = Add content flags
        mail = Mail { mBase = lname
                    , mglobalName = gname
                    , mFlags = flags
                    }
        in acceptPatch peer peerState gs mb gname operation $ S.insert mail
