module Sync ( syncAllMailBoxes
            , syncAllMailBoxesInit
            , syncMailBox
            ) where

import Data
import ExitCodes
import IOClasses
import IOData (withGlobalStateLocked, writeGlobalState_)
import LocalUpdates (new, update, addNewMail)
import Maildir (getMails, findMail)
import Match (getMailBoxes, mailBoxGuard, mailBoxGuard_)
import RemoteUpdates (anyPeer)
import RIO
import qualified RIO.HashMap as H
import qualified RIO.Text as T
import Utils (maybeReadBinaryFile, (>>=?), printErrorMessage)

syncAllMailBoxes :: (HasMailSyncConfig env, HasLogFunc env, HasMailRoot env)
                 => RIO env ()
syncAllMailBoxes = withGlobalStateLocked $ \gs -> do
  mbs <- getMailBoxes
  let syncOne = \gs' mb -> do
        logInfo . display . T.pack $ "Syncing MailBox: " ++ show mb
        syncMailBoxUnlocked gs' mb
  gs'' <- foldM syncOne gs mbs
  writeGlobalState_ gs''

syncAllMailBoxesInit :: (HasMailSyncConfig env, HasLogFunc env, HasMailRoot env)
                 => RIO env ()
syncAllMailBoxesInit = do
  whenM alreadyPeersPresent failAlreadyPeersPresent
  mbs <- getMailBoxes
  forM_ mbs $ \mb -> do
    logInfo . display . T.pack $ "Syncing MailBox: " ++ show mb
    syncMailBoxInit mb

alreadyPeersPresent :: (HasMailSyncConfig env) => RIO env Bool
alreadyPeersPresent = liftM (/= []) $ view mailSyncPeersL 

failAlreadyPeersPresent :: RIO env ()
failAlreadyPeersPresent = do
  printErrorMessage "There are already peers, refusing to perform initial operations"
  exitWith $ ExitFailure exitAlreadyPeers

syncMailBox :: (HasMailSyncConfig env, HasLogFunc env, HasMailRoot env)
            => String
            -> RIO env ()
syncMailBox mb = withGlobalStateLocked $ \gs ->
                  syncMailBoxUnlocked gs mb >>= writeGlobalState_

syncMailBoxUnlocked :: (HasMailSyncConfig env, HasLogFunc env, HasMailRoot env)
                    => GlobalState
                    -> String
                    -> RIO env GlobalState
syncMailBoxUnlocked gs mbName = mailBoxGuard gs mbName $ do
    let mb = createMailBox mbName
    maybeod <- getMails mb
    maybe (logWarn "Unknown mailbox" >> return gs) `flip` maybeod $ \ onDisk -> do
       let (gs', newmails) = update gs mb onDisk
       let readOneMail st mail = do
             maybecontent <- findMail mb mail >>=? maybeReadBinaryFile
             maybe (return st) `flip` maybecontent $ \ content -> do
               return $ addNewMail st mb (mail, mailContentFromByteString content)
       foldM readOneMail gs' newmails

syncMailBoxInit :: (HasMailSyncConfig env, HasLogFunc env, HasMailRoot env)
                => String
                -> RIO env ()
syncMailBoxInit mbName = mailBoxGuard_ mbName $
  withGlobalStateLocked $ \gs -> unless (anyPeer gs) $ do
    let mb = createMailBox mbName
    let gs' = gs { state = H.delete mb $ state gs }
    maybeod <- getMails mb
    maybe (logWarn "Unknown Mailbox") `flip` maybeod $ \onDisk ->
      writeGlobalState_ $ new gs' mb onDisk
