module Update (update, updateAll)
where

import RIO
import RIO.List (uncons)
import RIO.Process
import qualified RIO.Text as T

import Control.Monad.Trans.Except

import CliUtils (getPeer, requestFrom')
import Data
import Peer
import IOClasses
import IOData (slurpGlobalState)
import Utils


update :: (HasMailSyncConfig env, HasProcessContext env, HasLogFunc env)
       => String -> RIO env ()
update ident = do
  update' ident >>= either (logError . displayShow) (const $ return ())

requestToUs :: (HasProcessContext env, HasLogFunc env)
            => String -> [String] -> ExceptT String (RIO env) String
requestToUs cmd args = ExceptT $ do
  (exit, output) <- proc cmd args readProcessStdout
  return $ case exit of
    ExitFailure e -> Left $ "Could not get request: " ++ show e
    ExitSuccess -> mapBoth show T.unpack . decodeUtf8' $ toStrictBytes output


fromPeer :: (HasMailSyncConfig env, HasProcessContext env, HasLogFunc env)
         => String -> [String] -> String -> String -> RIO env (Either String ())
fromPeer cmd args request ident =  do
  let request' = words request
  let args' = args ++ request'
  (myExe, myArgs) <- view mailSyncMyCmdL
  proc cmd args' $ \pc1 -> do
    let pc = setStdout createPipe 
           $ setStdin closed
             pc1
    withProcessWait pc $ \p1 ->
      proc myExe (myArgs ++ ["peerToPeerProtocol", "applyMany", ident]) $ \pc2' -> do
        let pc2 = setStdin (useHandleOpen (getStdout p1))
                  pc2'
        withProcessWait pc2 $ \p2 -> do
          exitThere <- waitExitCode p1
          hClose (getStdout p1)
          case exitThere of
            ExitSuccess -> do
              exitHere <- waitExitCode p2
              case exitHere of
                ExitSuccess -> return $ Right ()
                ExitFailure e -> return . Left $ "Failed locally with exit code: " ++ show e ++ ", cmd was: " ++ show [myExe, "applyMany", ident]
            ExitFailure e -> do
              stopProcess p2
              return . Left $ "Failed remotely with exit code: " ++ show e


toPeer :: (HasMailSyncConfig env, HasProcessContext env, HasLogFunc env)
         => String -> [String] -> String -> String -> RIO env (Either String ())
toPeer cmd args request myIdent = do
  let request' = words request
  let args' = args ++ ["peerToPeerProtocol", "applyMany", myIdent]
  (myExe, myArgs) <- view mailSyncMyCmdL
  proc myExe (myArgs ++ request') $ \pc1' -> do
    let pc1 = setStdout createPipe
            $ setStdin closed
             pc1'
    withProcessWait pc1 $ \p1 ->
      proc cmd args' $ \pc2' -> do
        let pc2 = setStdin (useHandleOpen (getStdout p1))
                  pc2'
        withProcessWait pc2 $ \p2 -> do
          exitHere <- waitExitCode p1
          hClose (getStdout p1)
          case exitHere of
            ExitSuccess -> do
              exitThere <- waitExitCode p2
              case exitThere of
                ExitSuccess -> return $ Right ()
                ExitFailure e -> return . Left $ "Failed remotely with exit code: " ++ show e
            ExitFailure e -> do
              stopProcess p2
              return . Left $ "Failed locally with exit code: " ++ show e ++ ", cmd was: " ++ show (myExe:request')

update' :: (HasMailSyncConfig env, HasProcessContext env, HasLogFunc env)
        => String -> RIO env (Either String ())
update' ident = runExceptT $ do
  peer <- ExceptT $ maybeToRight "Did not find peer" <$> getPeer ident
  updatePeer peer

updatePeer :: (HasMailSyncConfig env, HasProcessContext env, HasLogFunc env)
           => Peer -> ExceptT String (RIO env) ()
updatePeer peer = do
  let ident = show $ pUUID peer
  path <- except . maybeToRight "No path to peer" $ pPath peer 
  (cmd, args) <- except . maybeToRight "Empty path" $ uncons path
  requestFromUs <- ExceptT $ maybeToRight "Did not find peer" <$> requestFrom' ident
  gs <- lift slurpGlobalState
  request' <- requestToUs cmd $ args ++ ["peerToPeerProtocol", "requestFrom", show $ myUUID gs]
  logInfo $ displayShow requestFromUs
  logInfo $ displayShow request'
  toPeerAsync <- lift . async $ toPeer cmd args request' $ (show .  myUUID) gs
  fromPeerAsync <- lift . async $ fromPeer cmd args requestFromUs ident
  (fromPeer', toPeer') <- lift $ waitBoth fromPeerAsync toPeerAsync
  except $ fromPeer' >> toPeer'

updateAll :: (HasMailSyncConfig env, HasProcessContext env, HasLogFunc env)
          => RIO env ()
updateAll = do
  peers <- view mailSyncPeersL
  let callablePeers = filter (isJust . pPath) peers
  callResults <- mapM (runExceptT . updatePeer) callablePeers
  let errors = filter (RIO.isLeft . snd) $ zip callablePeers callResults
  when (not $ null errors) . logError . display . T.pack
     $ "Failures calling peers: " ++ show errors
