#!/bin/env bats

load '../../bats-support/load'
load '../../bats-file/load'

function setup {
    STACK_INSTALL_ROOT=$(cd mail-sync && stack path --local-install-root)
    MAIL_SYNC="${STACK_INSTALL_ROOT}/bin/mail-sync"
    TEST_TMP_DIR="$(temp_make)"
    MAIL_SYNC_1="${MAIL_SYNC} --config ${TEST_TMP_DIR}/test_config"
    MAIL_SYNC_2="${MAIL_SYNC} --config ${TEST_TMP_DIR}/test_config2"
    mkdir "${TEST_TMP_DIR}/sync"
    mkdir "${TEST_TMP_DIR}/sync2"
    mkdir "${TEST_TMP_DIR}/mail"
    mkdir "${TEST_TMP_DIR}/mail2"
    cp test/data/test_config "${TEST_TMP_DIR}"
    cp test/data/test_config "${TEST_TMP_DIR}/test_config2"
    ed "${TEST_TMP_DIR}/test_config" << EOF
s|ROOT|${TEST_TMP_DIR}/mail
s|SYNC|${TEST_TMP_DIR}/sync
s|CMD|("${MAIL_SYNC}", ["--config", "${TEST_TMP_DIR}/test_config"])
w
q
EOF
    ed "${TEST_TMP_DIR}/test_config2" << EOF
s|ROOT|${TEST_TMP_DIR}/mail2
s|SYNC|${TEST_TMP_DIR}/sync2
s|CMD|("${MAIL_SYNC}", ["--config", "${TEST_TMP_DIR}/test_config2"])
w
q
EOF
    cp -r test/data/testMails "${TEST_TMP_DIR}/mail"
    mkdir -p "${TEST_TMP_DIR}/mail/testMails/new"
    mkdir -p "${TEST_TMP_DIR}/mail/testMails/tmp"
}

function teardown {
    temp_del "$TEST_TMP_DIR"
}

@test "oneSync" {
  run sh -c "${MAIL_SYNC_1} setup init test1"
  echo "$output"
  [ "$status" -eq 0 ]
  assert_file_exist ${TEST_TMP_DIR}/sync/state.global
  run sh -c "${MAIL_SYNC_1} syncOne testMails"
  [ "$status" -eq 0 ]
  assert_file_exist ${TEST_TMP_DIR}/sync/state.global

  run sh -c "${MAIL_SYNC_2} setup init test2"
  [ "$status" -eq 0 ]
  UUID1=`${MAIL_SYNC_1} setup printUUID`
  UUID2=`${MAIL_SYNC_2} setup printUUID`
  run sh -c "${MAIL_SYNC_1} peerToPeerProtocol sendSince ${UUID2} 0 > ${TEST_TMP_DIR}/output1"
  [ "$status" -eq 0 ]
  run sh -c "${MAIL_SYNC_2} peerToPeerProtocol applyMany ${UUID1} < ${TEST_TMP_DIR}/output1"
  [ "$status" -eq 0 ]
  run sh -c "ls -l ${TEST_TMP_DIR}/mail2/testMails/cur/ | wc -l"
  [ "${lines[0]}" -eq 13 ] 

  # Replaying the already known patches should have no effect
  run sh -c "${MAIL_SYNC_2} applyMany ${UUID1} < ${TEST_TMP_DIR}/output1"
  run sh -c "ls -l ${TEST_TMP_DIR}/mail2/testMails/cur/ | wc -l"
  echo "Lines after trying to aplly same patch again:" ${lines[0]}
  [ "${lines[0]}" -eq 13 ]
}

@test "global State read-write consistency" {
  run sh -c "${MAIL_SYNC_1} setup init test1"
  run sh -c "${MAIL_SYNC_1} syncOne testMails"
  run sh -c "${MAIL_SYNC_1} debugState showGlobalState | ${MAIL_SYNC_1} debugState encodeGlobalState >  ${TEST_TMP_DIR}/gsnew"
  ls -al "${TEST_TMP_DIR}/sync/state.global" "${TEST_TMP_DIR}/gsnew"
  cmp "${TEST_TMP_DIR}/sync/state.global" "${TEST_TMP_DIR}/gsnew"
}

@test "nothing to sync if there are no changes" {
  run sh -c "${MAIL_SYNC_1} setup init test1"
  run sh -c "${MAIL_SYNC_1} syncOne testMails"
  [ "$status" -eq 0 ]
  run sh -c "${MAIL_SYNC_2} setup init test2"
  [ "$status" -eq 0 ]
  UUID1=`${MAIL_SYNC_1} setup printUUID`
  UUID2=`${MAIL_SYNC_2} setup printUUID`
  run sh -c "${MAIL_SYNC_1} peerToPeerProtocol sendSince ${UUID2} 0 > ${TEST_TMP_DIR}/output1"
  [ "$status" -eq 0 ]
  run sh -c "${MAIL_SYNC_2} peerToPeerProtocol applyMany ${UUID1} < ${TEST_TMP_DIR}/output1"
  [ "$status" -eq 0 ]
  run sh -c "${MAIL_SYNC_2} peerToPeerProtocol sendSince ${UUID1} 0 > ${TEST_TMP_DIR}/output2"
  [ ! -s ${TEST_TMP_DIR}/output2 ] 
}

@test "no new updates after applying known ones" {
  # Keep one mail back for later
  NEW_MAIL=`ls ${TEST_TMP_DIR}/mail/testMails/cur/1* | head -1`
  mv ${NEW_MAIL} ${TEST_TMP_DIR}/new_mail

  # Set up and do one sync
  run sh -c "${MAIL_SYNC_1} setup init test1"
  run sh -c "${MAIL_SYNC_1} syncOne testMails"
  [ "$status" -eq 0 ]
  SIZE_STATE=`cat ${TEST_TMP_DIR}/sync/state.global | wc -c`
  run sh -c "${MAIL_SYNC_2} setup init test2"
  [ "$status" -eq 0 ]
  UUID1=`${MAIL_SYNC_1} setup printUUID`
  UUID2=`${MAIL_SYNC_2} setup printUUID`

  REQUEST=`${MAIL_SYNC_2} peerToPeerProtocol requestFrom ${UUID1}`

  echo
  echo ${REQUEST}
  echo

  run sh -c "${MAIL_SYNC_1} ${REQUEST} > ${TEST_TMP_DIR}/output1"
  [ "$status" -eq 0 ]
  run sh -c "${MAIL_SYNC_2} peerToPeerProtocol applyMany ${UUID1} < ${TEST_TMP_DIR}/output1"
  [ "$status" -eq 0 ]


  # Now that the peer 1 is aware of peer 2, but only knows that peer 2 knowns at least
  # nothing, purging should be a no-op.
  SIZE_STATE_BEFORE_NOOP_PURGE=`cat ${TEST_TMP_DIR}/sync/state.global | wc -c`
  run sh -c "${MAIL_SYNC_1} purge"
  [ "$status" -eq 0 ]
  SIZE_STATE_AFTER_NOOP_PURGE=`cat ${TEST_TMP_DIR}/sync/state.global | wc -c`
  echo "Size before and after no-op purge: ${SIZE_STATE_BEFORE_NOOP_PURGE} and ${SIZE_STATE_AFTER_NOOP_PURGE}"
  [ ${SIZE_STATE_BEFORE_NOOP_PURGE}  -eq ${SIZE_STATE_AFTER_NOOP_PURGE} ]

  # As all changes are included, requesting updates again should not provide anything new
  NEW_REQUEST=`${MAIL_SYNC_2} peerToPeerProtocol requestFrom ${UUID1}`

  echo
  echo ${NEW_REQUEST}
  echo

  run sh -c "${MAIL_SYNC_1} ${NEW_REQUEST} > ${TEST_TMP_DIR}/output1new"
  [ "$status" -eq 0 ]
  [ ! -s ${TEST_TMP_DIR}/output1new ]

  # Now, add the mail that we kept back
  mv ${TEST_TMP_DIR}/new_mail ${NEW_MAIL}
  # Update on peer 1
  run sh -c "${MAIL_SYNC_1} syncOne testMails"
  [ "$status" -eq 0 ]
  # Now, requesting updates from peer one should contain updates
  run sh -c "${MAIL_SYNC_1} ${NEW_REQUEST} > ${TEST_TMP_DIR}/output1new2"
  [ "$status" -eq 0 ]
  [ -s ${TEST_TMP_DIR}/output1new2 ]

  # After seeing the new request, peer 1 should be able to meaningfully purge
  run sh -c "${MAIL_SYNC_1} purge"
  [ "$status" -eq 0 ]
  SIZE_STATE_AFTER_PURGE=`cat ${TEST_TMP_DIR}/sync/state.global | wc -c`

  echo
  echo Purged from ${SIZE_STATE} bytes to ${SIZE_STATE_AFTER_PURGE} bytes
  echo

  [ ${SIZE_STATE_AFTER_PURGE} -lt ${SIZE_STATE} ]
}
